@@ -7,13 +7,16 @@ import qualified Control.Monad as Monad
 import qualified Data.Char as Char
 import qualified Data.List as List
 import qualified Data.Maybe as Maybe
+import qualified Data.Monoid as Monoid
 import qualified Data.Ord as Ord
+import qualified Data.Version as Version
+import qualified Data.Word as Word
 import qualified Text.ParserCombinators.ReadP as ReadP
-import qualified Text.Read as Read
 
 -- $setup
+-- >>> import Control.Applicative
 -- >>> import Lens.Micro
--- >>> import Lens.Micro.Extras
+-- >>> let view l x = x ^. l
 
 -- * Public
 
@@ -27,14 +30,14 @@ import qualified Text.Read as Read
 --
 -- Use 'parseVersion' to create versions.
 data Version = Version
-  { versionMajor :: Word
-  , versionMinor :: Word
-  , versionPatch :: Word
+  { versionMajor :: Word.Word64
+  , versionMinor :: Word.Word64
+  , versionPatch :: Word.Word64
   , versionPreReleases :: [PreRelease]
   , versionBuilds :: [Build]
   } deriving (Eq, Show)
 
--- | In general, 'Version's compare in the way that you would expect. First the
+-- | In general, versions compare in the way that you would expect. First the
 -- major version numbers are compared, then the minors, then the patches.
 --
 -- >>> compare <$> parseVersion "1.2.3" <*> parseVersion "2.0.0"
@@ -69,7 +72,7 @@ data Version = Version
 -- >>> (==) <$> parseVersion "1.2.3+a" <*> parseVersion "1.2.3+b"
 -- Just False
 instance Ord Version where
-  compare x y = mconcat
+  compare x y = Monoid.mconcat
     [ Ord.comparing versionMajor x y
     , Ord.comparing versionMinor x y
     , Ord.comparing versionPatch x y
@@ -93,7 +96,7 @@ instance Ord Version where
 --
 -- Use 'parsePreRelease' to create pre-releases.
 data PreRelease
-  = PreReleaseNumeric Word
+  = PreReleaseNumeric Word.Word64
   | PreReleaseTextual String
   deriving (Eq, Show)
 
@@ -131,8 +134,8 @@ newtype Build = Build String deriving (Eq, Show)
 
 -- | Constrains allowable version numbers.
 --
--- Use 'parseConstraint' to create constraints and 'satisfies' to see if a
--- version number satisfies a constraint.
+-- Use 'parseConstraint' to create constraints and 'satisfiesConstraint' to see
+-- if a version number satisfies a constraint.
 data Constraint
   = ConstraintOperator Operator Version
   | ConstraintHyphen Version Version
@@ -143,12 +146,12 @@ data Constraint
 
 -- | Makes a new version number.
 --
--- >>> makeVersion 0 0 0 [] []
--- Version {versionMajor = 0, versionMinor = 0, versionPatch = 0, versionPreReleases = [], versionBuilds = []}
+-- >>> makeVersion 1 2 3 [unsafeParsePreRelease "pre"] [unsafeParseBuild "build"]
+-- Version {versionMajor = 1, versionMinor = 2, versionPatch = 3, versionPreReleases = [PreReleaseTextual "pre"], versionBuilds = [Build "build"]}
 --
 -- This can be a useful alternative to 'parseVersion' if you want a total way
 -- to create a version.
-makeVersion :: Word -> Word -> Word -> [PreRelease] -> [Build] -> Version
+makeVersion :: Word.Word64 -> Word.Word64 -> Word.Word64 -> [PreRelease] -> [Build] -> Version
 makeVersion major minor patch preReleases builds = Version
   { versionMajor = major
   , versionMinor = minor
@@ -164,8 +167,8 @@ makeVersion major minor patch preReleases builds = Version
 initialVersion :: Version
 initialVersion = makeVersion 0 0 0 [] []
 
--- | Attempts to parse a version. This parser follows [SemVer's
--- BNF](https://github.com/mojombo/semver/blob/eb9aac5/semver.md#backusnaur-form-grammar-for-valid-semver-versions).
+-- | Attempts to parse a version. This parser follows
+-- [SemVer's BNF](https://github.com/mojombo/semver/blob/eb9aac5/semver.md#backusnaur-form-grammar-for-valid-semver-versions).
 --
 -- >>> parseVersion "1.2.3-p.4+b.5"
 -- Just (Version {versionMajor = 1, versionMinor = 2, versionPatch = 3, versionPreReleases = [PreReleaseTextual "p",PreReleaseNumeric 4], versionBuilds = [Build "b",Build "5"]})
@@ -220,10 +223,8 @@ parsePreRelease s = parse preReleaseP s
 parseBuild :: String -> Maybe Build
 parseBuild s = parse buildP s
 
--- | Attempts to parse a constraint. This parser follows [npm's
--- BNF](https://github.com/npm/npm/blob/d081cc6/doc/misc/semver.md#range-grammar),
--- except that neither the so-called "x-ranges" nor partial version numbers are
--- not supported. So you cannot use @1.2.x@ or @>1.2@ as version constraints.
+-- | Attempts to parse a constraint. This parser mostly follows
+-- [npm's BNF](https://github.com/npm/npm/blob/d081cc6/doc/misc/semver.md#range-grammar).
 --
 -- >>> parseConstraint ">1.2.3"
 -- Just (ConstraintOperator OperatorGT (Version {versionMajor = 1, versionMinor = 2, versionPatch = 3, versionPreReleases = [], versionBuilds = []}))
@@ -232,6 +233,19 @@ parseBuild s = parse buildP s
 --
 -- >>> parseConstraint "wrong"
 -- Nothing
+--
+-- The two departures from npm's BNF are that x-ranges cannot be used with
+-- other operators and partial version numbers are not allowed.
+--
+-- >>> parseConstraint "1.2.x"
+-- Just (ConstraintWildcard (WildcardPatch 1 2))
+-- >>> parseConstraint ">=1.2.x"
+-- Nothing
+--
+-- >>> parseConstraint "1.2"
+-- Nothing
+-- >>> parseConstraint ">=1.2"
+-- Nothing
 parseConstraint :: String -> Maybe Constraint
 parseConstraint s = parse constraintsP s
 
@@ -308,7 +322,7 @@ unsafeParseConstraint s = case parseConstraint s of
 -- >>> renderVersion <$> parseVersion "1.2.3-p.4+b.5"
 -- Just "1.2.3-p.4+b.5"
 renderVersion :: Version -> String
-renderVersion v = mconcat
+renderVersion v = concat
   [ show (versionMajor v)
   , "."
   , show (versionMinor v)
@@ -386,6 +400,57 @@ isUnstable v = versionMajor v == 0
 isStable :: Version -> Bool
 isStable v = not (isUnstable v)
 
+-- | Converts from a 'Version.Version' from the @base@ package.
+--
+-- >>> renderVersion . fromBaseVersion $ Version.makeVersion [1, 2, 3]
+-- "1.2.3"
+--
+-- Missing version components are set to zero.
+--
+-- >>> renderVersion . fromBaseVersion $ Version.makeVersion []
+-- "0.0.0"
+-- >>> renderVersion . fromBaseVersion $ Version.makeVersion [1]
+-- "1.0.0"
+-- >>> renderVersion . fromBaseVersion $ Version.makeVersion [1, 2]
+-- "1.2.0"
+--
+-- Extra version components are ignored.
+--
+-- >>> renderVersion . fromBaseVersion $ Version.makeVersion [1, 2, 3, 4]
+-- "1.2.3"
+--
+-- Tags are ignored.
+--
+-- >>> renderVersion . fromBaseVersion $ Version.Version [] ["ignored"]
+-- "0.0.0"
+fromBaseVersion :: Version.Version -> Version
+fromBaseVersion v = case Version.versionBranch v of
+  (m : n : p : _) -> mkV (fromIntegral m) (fromIntegral n) (fromIntegral p)
+  (m : n : _) -> mkV (fromIntegral m) (fromIntegral n) 0
+  (m : _) -> mkV (fromIntegral m) 0 0
+  _ -> mkV 0 0 0
+
+-- | Converts to a 'Version.Version' from the @base@ package.
+--
+-- >>> toBaseVersion <$> parseVersion "1.2.3"
+-- Just (Version {versionBranch = [1,2,3], versionTags = []})
+--
+-- Pre-releases and builds are converted to tags.
+--
+-- >>> toBaseVersion <$> parseVersion "1.2.3-pre+build"
+-- Just (Version {versionBranch = [1,2,3], versionTags = ["pre","build"]})
+toBaseVersion :: Version -> Version.Version
+toBaseVersion v = Version.Version
+  (map fromIntegral
+    [ versionMajor v
+    , versionMinor v
+    , versionPatch v
+    ])
+  (concat
+    [ map renderPreRelease (versionPreReleases v)
+    , map renderBuild (versionBuilds v)
+    ])
+
 -- | Increments the major version number.
 --
 -- >>> bumpMajor <$> parseVersion "0.0.0"
@@ -449,32 +514,10 @@ bumpPatch v = makeVersion
 
 -- | Returns 'True' if the version satisfies the constraint, 'False' otherwise.
 --
--- >>> satisfies <$> parseVersion "1.2.3" <*> parseConstraint ">1.2.0"
+-- >>> satisfiesConstraint <$> parseConstraint ">1.2.0" <*> parseVersion "1.2.3"
 -- Just True
-satisfies :: Version -> Constraint -> Bool
-satisfies v c = case c of
-  ConstraintOperator o w -> case o of
-    OperatorLT -> v < w
-    OperatorLE -> v <= w
-    OperatorEQ -> compare v w == EQ
-    OperatorGE -> v >= w
-    OperatorGT -> v > w
-    OperatorTilde -> (v >= w) && (v < makeVersion (versionMajor w) (versionMinor w + 1) 0 [] [])
-    OperatorCaret ->
-      let u =
-            if versionMajor w == 0
-            then if versionMinor w == 0
-            then makeVersion (versionMajor w) (versionMinor w) (versionPatch w + 1) [] []
-            else makeVersion (versionMajor w) (versionMinor w + 1) 0 [] []
-            else makeVersion (versionMajor w + 1) 0 0 [] []
-      in (v >= w) && (v < u)
-  ConstraintHyphen l r -> (l <= v) && (v <= r)
-  ConstraintWildcard w -> case w of
-    WildcardMajor -> True
-    WildcardMinor m -> (makeVersion m 0 0 [] [] <= v) && (v < makeVersion (m + 1) 0 0 [] [])
-    WildcardPatch m n -> (makeVersion m n 0 [] [] <= v) && (v < makeVersion m (n + 1) 0 [] [])
-  ConstraintAnd l r -> satisfies v l && satisfies v r
-  ConstraintOr l r -> satisfies v l || satisfies v r
+satisfiesConstraint :: Constraint -> Version -> Bool
+satisfiesConstraint c v = satisfiesSC (toSC c) v
 
 -- | Focuses on the major version number.
 --
@@ -482,7 +525,7 @@ satisfies v c = case c of
 -- Just 1
 -- >>> set majorLens 4 <$> parseVersion "1.2.3"
 -- Just (Version {versionMajor = 4, versionMinor = 2, versionPatch = 3, versionPreReleases = [], versionBuilds = []})
-majorLens :: Functor f => (Word -> f Word) -> Version -> f Version
+majorLens :: Functor f => (Word.Word64 -> f Word.Word64) -> Version -> f Version
 majorLens f v = fmap
   (\ m -> v { versionMajor = m })
   (f (versionMajor v))
@@ -493,7 +536,7 @@ majorLens f v = fmap
 -- Just 2
 -- >>> set minorLens 4 <$> parseVersion "1.2.3"
 -- Just (Version {versionMajor = 1, versionMinor = 4, versionPatch = 3, versionPreReleases = [], versionBuilds = []})
-minorLens :: Functor f => (Word -> f Word) -> Version -> f Version
+minorLens :: Functor f => (Word.Word64 -> f Word.Word64) -> Version -> f Version
 minorLens f v = fmap
   (\ n -> v { versionMinor = n })
   (f (versionMinor v))
@@ -504,7 +547,7 @@ minorLens f v = fmap
 -- Just 3
 -- >>> set patchLens 4 <$> parseVersion "1.2.3"
 -- Just (Version {versionMajor = 1, versionMinor = 2, versionPatch = 4, versionPreReleases = [], versionBuilds = []})
-patchLens :: Functor f => (Word -> f Word) -> Version -> f Version
+patchLens :: Functor f => (Word.Word64 -> f Word.Word64) -> Version -> f Version
 patchLens f v = fmap
   (\ p -> v { versionPatch = p })
   (f (versionPatch v))
@@ -547,8 +590,8 @@ data Operator
 
 data Wildcard
   = WildcardMajor
-  | WildcardMinor Word
-  | WildcardPatch Word Word
+  | WildcardMinor Word.Word64
+  | WildcardPatch Word.Word64 Word.Word64
   deriving (Eq, Show)
 
 -- ** Constructors
@@ -657,7 +700,7 @@ versionP = do
   patch <- numberP
   preReleases <- preReleasesP
   builds <- buildsP
-  pure (makeVersion major minor patch preReleases builds)
+  return (makeVersion major minor patch preReleases builds)
 
 preReleasesP :: ReadP.ReadP [PreRelease]
 preReleasesP = ReadP.option [] (do
@@ -670,14 +713,14 @@ preReleaseP = preReleaseNumberP ReadP.<++ preReleaseStringP
 preReleaseNumberP :: ReadP.ReadP PreRelease
 preReleaseNumberP = do
   n <- numberP
-  pure (PreReleaseNumeric n)
+  return (PreReleaseNumeric n)
 
 preReleaseStringP :: ReadP.ReadP PreRelease
 preReleaseStringP = do
   s <- ReadP.munch1 isIdentifier
   if all Char.isDigit s
     then ReadP.pfail
-    else pure (PreReleaseTextual s)
+    else return (PreReleaseTextual s)
 
 buildsP :: ReadP.ReadP [Build]
 buildsP = ReadP.option [] (do
@@ -687,21 +730,21 @@ buildsP = ReadP.option [] (do
 buildP :: ReadP.ReadP Build
 buildP = do
   b <- ReadP.munch1 isIdentifier
-  pure (Build b)
+  return (Build b)
 
-numberP :: ReadP.ReadP Word
-numberP = nonZeroP ReadP.<++ zeroP
+numberP :: ReadP.ReadP Word.Word64
+numberP = zeroP ReadP.<++ nonZeroP
 
-zeroP :: ReadP.ReadP Word
+zeroP :: ReadP.ReadP Word.Word64
 zeroP = do
   Monad.void (ReadP.char '0')
-  pure 0
+  return 0
 
-nonZeroP :: ReadP.ReadP Word
+nonZeroP :: ReadP.ReadP Word.Word64
 nonZeroP = do
   x <- ReadP.satisfy isAsciiDigitNonZero
   ys <- ReadP.munch Char.isDigit
-  case Read.readMaybe (x : ys) of
+  case toWord64 (stringToIntegral (x : ys)) of
     Nothing -> ReadP.pfail
     Just n -> pure n
 
@@ -710,27 +753,28 @@ constraintsP = do
   spacesP
   cs <- ReadP.sepBy1 constraintP orP
   spacesP
-  pure (foldr1 constraintOr cs)
+  return (foldr1 constraintOr cs)
 
 constraintP :: ReadP.ReadP Constraint
 constraintP = do
   cs <- ReadP.sepBy1 simpleP spaces1P
-  pure (foldr1 constraintAnd cs)
+  return (foldr1 constraintAnd cs)
 
 hyphenatedP :: ReadP.ReadP Constraint
 hyphenatedP = do
   v <- versionP
   hyphenP
   w <- versionP
-  pure (constraintHyphen v w)
+  return (constraintHyphen v w)
 
 simpleP :: ReadP.ReadP Constraint
 simpleP = ReadP.choice [hyphenatedP, wildcardConstraintP, primitiveP]
 
 wildcardConstraintP :: ReadP.ReadP Constraint
 wildcardConstraintP = do
+  ReadP.optional (ReadP.char '=')
   w <- wildcardP
-  pure (ConstraintWildcard w)
+  return (ConstraintWildcard w)
 
 wildcardP :: ReadP.ReadP Wildcard
 wildcardP = ReadP.choice [wildcardPatchP, wildcardMinorP, wildcardMajorP]
@@ -742,7 +786,7 @@ wildcardPatchP = do
   n <- numberP
   Monad.void (ReadP.char '.')
   Monad.void (ReadP.satisfy isWildcard)
-  pure (WildcardPatch m n)
+  return (WildcardPatch m n)
 
 wildcardMinorP :: ReadP.ReadP Wildcard
 wildcardMinorP = do
@@ -751,7 +795,7 @@ wildcardMinorP = do
   Monad.void (ReadP.satisfy isWildcard)
   Monad.void (ReadP.char '.')
   Monad.void (ReadP.satisfy isWildcard)
-  pure (WildcardMinor m)
+  return (WildcardMinor m)
 
 wildcardMajorP :: ReadP.ReadP Wildcard
 wildcardMajorP = do
@@ -760,25 +804,25 @@ wildcardMajorP = do
   Monad.void (ReadP.satisfy isWildcard)
   Monad.void (ReadP.char '.')
   Monad.void (ReadP.satisfy isWildcard)
-  pure WildcardMajor
+  return WildcardMajor
 
 primitiveP :: ReadP.ReadP Constraint
 primitiveP = do
   o <- operatorP
   spacesP
   v <- versionP
-  pure (ConstraintOperator o v)
+  return (ConstraintOperator o v)
 
 operatorP :: ReadP.ReadP Operator
 operatorP = ReadP.choice
-  [ ReadP.string "<=" *> pure OperatorLE
-  , ReadP.string ">=" *> pure OperatorGE
-  , ReadP.char '<' *> pure OperatorLT
-  , ReadP.char '>' *> pure OperatorGT
-  , ReadP.char '=' *> pure OperatorEQ
-  , ReadP.char '^' *> pure OperatorCaret
-  , ReadP.char '~' *> pure OperatorTilde
-  , pure OperatorEQ
+  [ ReadP.string "<=" >> return OperatorLE
+  , ReadP.string ">=" >> return OperatorGE
+  , ReadP.char '<' >> return OperatorLT
+  , ReadP.char '>' >> return OperatorGT
+  , ReadP.char '=' >> return OperatorEQ
+  , ReadP.char '^' >> return OperatorCaret
+  , ReadP.char '~' >> return OperatorTilde
+  , return OperatorEQ
   ]
 
 hyphenP :: ReadP.ReadP ()
@@ -799,15 +843,6 @@ spaces1P = Monad.void (ReadP.munch1 (== ' '))
 spacesP :: ReadP.ReadP ()
 spacesP = Monad.void (ReadP.munch (== ' '))
 
--- *** Helpers
-
-parse :: ReadP.ReadP a -> String -> Maybe a
-parse p s =
-  let p' = ReadP.readP_to_S p
-  in Maybe.listToMaybe (do
-    (x, "") <- p' s
-    pure x)
-
 -- ** Rendering
 
 renderPreReleases :: [PreRelease] -> String
@@ -833,3 +868,78 @@ isIdentifier c = (Char.isAscii c && Char.isAlphaNum c) || (c == '-')
 
 isWildcard :: Char -> Bool
 isWildcard c = (c == 'x') || (c == '*') || (c == 'X')
+
+parse :: ReadP.ReadP a -> String -> Maybe a
+parse p s =
+  let p' = ReadP.readP_to_S p
+  in Maybe.listToMaybe (do
+    (x, "") <- p' s
+    return x)
+
+stringToIntegral :: Integral a => String -> a
+stringToIntegral s = foldl
+  (\ n d -> (n * 10) + (fromIntegral (fromEnum d) - 48)) 0 s
+
+toWord64 :: Integer -> Maybe Word.Word64
+toWord64 n =
+  if n < 0
+  then Nothing
+  else if n > fromIntegral (maxBound :: Word.Word64)
+  then Nothing
+  else Just (fromIntegral n)
+
+-- * Simple constraints
+-- | Simple constraints are just as expressive as 'Constraint's, but they are
+-- easier to reason about. You can think of them as the desugared version of
+-- 'Constraint's.
+
+data SimpleConstraint
+  = SCLT Version
+  | SCEQ Version
+  | SCGT Version
+  | SCAnd SimpleConstraint SimpleConstraint
+  | SCOr SimpleConstraint SimpleConstraint
+  deriving (Eq, Show)
+
+mkV :: Word.Word64 -> Word.Word64 -> Word.Word64 -> Version
+mkV m n p = makeVersion m n p [] []
+
+satisfiesSC :: SimpleConstraint -> Version -> Bool
+satisfiesSC c v = case c of
+  SCLT u -> v < u
+  -- This uses `compare` rather than `==` to ignore build metadata.
+  SCEQ u -> compare v u == EQ
+  SCGT u -> v > u
+  SCAnd l r -> satisfiesSC l v && satisfiesSC r v
+  SCOr l r -> satisfiesSC l v || satisfiesSC r v
+
+scLE :: Version -> SimpleConstraint
+scLE v = SCOr (SCLT v) (SCEQ v)
+
+scGE :: Version -> SimpleConstraint
+scGE v = SCOr (SCGT v) (SCEQ v)
+
+toSC :: Constraint -> SimpleConstraint
+toSC c = case c of
+  ConstraintOperator o v -> case o of
+    OperatorLT -> SCLT v
+    OperatorLE -> scLE v
+    OperatorEQ -> SCEQ v
+    OperatorGE -> scGE v
+    OperatorGT -> SCGT v
+    OperatorTilde -> SCAnd
+      (scGE v)
+      (SCLT (mkV (versionMajor v) (versionMinor v + 1) 0))
+    OperatorCaret -> SCAnd
+      (scGE v)
+      (SCLT (case (versionMajor v, versionMinor v, versionPatch v) of
+        (0, 0, p) -> mkV 0 0 (p + 1)
+        (0, n, _) -> mkV 0 (n + 1) 0
+        (m, _, _) -> mkV (m + 1) 0 0))
+  ConstraintHyphen l h -> SCAnd (scGE l) (scLE h)
+  ConstraintWildcard w -> case w of
+    WildcardMajor -> scGE initialVersion
+    WildcardMinor m -> SCAnd (scGE (mkV m 0 0)) (SCLT (mkV (m + 1) 0 0))
+    WildcardPatch m n -> SCAnd (scGE (mkV m n 0)) (SCLT (mkV m (n + 1) 0))
+  ConstraintAnd l r -> SCAnd (toSC l) (toSC r)
+  ConstraintOr l r -> SCOr (toSC l) (toSC r)
